#!/usr/bin/env python

"""
Input context common functionality.

Copyright (C) 2018, 2019 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from moinformat.utils.file import readfile
from os.path import split
import codecs

class Input:

    "A common input context abstraction."

    default_encoding = "utf-8"

    def __init__(self, metadata):

        "Initialise the input context with the given 'metadata'."

        self.metadata = metadata

        # Obtain essential metadata.

        self.encoding = metadata.get("input_encoding", self.default_encoding)

    def all(self):

        "Return all pages in the context."

        return []

    def all_attachments(self):

        "Return all attachment filenames in the context."

        return []

    def get_attachments(self, pagename):

        "Return all attachment filenames for the given 'pagename'."

        return []

    # Page characteristics.

    def parent(self, pagename):

        "Return the parent of 'pagename'."

        return "/" in pagename and pagename.rsplit("/", 1)[0] or None

    def subpages(self, pagename):

        "Return the subpages of 'pagename'."

        return []

    # Page access methods.

    def readfile(self, filename, encoding=None):

        """
        Return the contents of the file having the given 'filename' and optional
        'encoding'. This implementation treats 'filename' as a path.
        """

        return self.readpath(filename, encoding)

    def readpage(self, pagename, encoding=None):

        """
        Return the contents of the file having the given 'pagename' and optional
        'encoding'.
        """

        return self.readfile(self.to_filename(pagename), encoding)

    # Input methods.

    def readpath(self, filename, encoding=None):

        """
        Return the contents of the file having the given 'filename'. If the
        optional 'encoding' is specified, override the general encoding.
        """

        return readfile(filename, encoding or self.encoding)

    # Name translation methods.

    def to_filename(self, pagename):

        "Return the filename corresponding to 'pagename'."

        return pagename

    def to_pagename(self, filename):

        "Return the pagename corresponding to 'filename'."

        # Take the leafname as the pagename from an arbitrary filename.

        return split(filename)[-1]

# vim: tabstop=4 expandtab shiftwidth=4
