#!/usr/bin/env python

"""
Directory input context.

Copyright (C) 2018, 2019 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from moinformat.input.common import Input
from moinformat.utils.directory import Directory
from os.path import join, sep

class DirectoryInput(Input):

    "A directory input context."

    name = "directory"

    def __init__(self, metadata):

        "Initialise the context with the given 'metadata'."

        if not metadata.has_key("input_filename"):
            raise ValueError, metadata

        Input.__init__(self, metadata)
        self.dir = Directory(metadata.get("input_filename"))

        # Support an encoding of the level separator for the filesystem.
        # Where it is the same as the directory separator, documents are stored
        # using nested directories, not as a flat list.

        self.level_sep = metadata.get("input_separator", sep)

        # Search recursively in nested directories for pages.

        self.nested = self.level_sep == sep

        # Support attachments directories.

        self.attachments_dir = metadata.get("attachments")

        # Support a common attachments directory.

        self.common_attachments = metadata.get("common_attachments")

    def _get_attachments_dir(self, pagename):

        "Return the attachments directory for 'pagename'."

        if self.common_attachments:
            return join(self.dir.filename, self.attachments_dir)
        else:
            return join(self.dir.filename, self.attachments_dir, pagename)

    def all(self):

        "Return all pages in the context."

        # Ignore dotfiles.

        return map(self.to_pagename, self.dir.select_files("[!.]*", self.nested))

    def all_attachments(self):

        "Return all attachment filenames in the context."

        return self.dir.select_files(join(self.attachments_dir, "*"), True)

    def get_attachments(self, pagename):

        """
        Return all attachment filenames for the given 'pagename'. Each filename
        is relative to the appropriate attachment directory.
        """

        return Directory(self._get_attachments_dir(pagename)).select_files("*")

    # Page characteristics.

    def subpage_filenames(self, pagename):

        "Return the subpage filenames of 'pagename'."

        pattern = self.to_filename("%s%s*" % (pagename, self.level_sep))
        return self.dir.select_files(pattern, self.nested)

    def subpages(self, pagename):

        "Return the subpages of 'pagename'."

        return map(self.to_pagename, self.subpage_filenames(pagename))

    # Page access methods.

    def readfile(self, filename, encoding=None):

        """
        Return the contents of the file having the given 'filename' and optional
        'encoding'.
        """

        return self.readpath(self.dir.get_filename(filename), encoding)

    # Convenience methods.

    def get_attachment_filename(self, pagename, filename):

        """
        Return the full path of an attachment file for the given 'pagename'
        having the given 'filename'.
        """

        if not pagename:
            return None

        return self.dir.get_filename(join(self._get_attachments_dir(pagename),
                                          filename))

    # NOTE: Translation methods should encode filenames appropriately.

    def to_filename(self, pagename):

        "Return the filename corresponding to 'pagename'."

        if sep == self.level_sep:
            return pagename
        else:
            return self.level_sep.join(pagename.split("/"))

    def to_pagename(self, filename):

        "Return the pagename corresponding to 'filename'."

        if sep == self.level_sep:
            return filename
        else:
            return "/".join(filename.split(self.level_sep))

input = DirectoryInput

# vim: tabstop=4 expandtab shiftwidth=4
