#!/usr/bin/env python

"""
Common linking scheme functionality.

Copyright (C) 2018, 2019 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

class Link:

    "A link abstraction."

    def __init__(self, target, label, link_target=None):

        """
        Initialise the link with the given 'target' and 'label' and
        'link_target' object.
        """

        self.target = target
        self.label = label
        self.link_target = link_target

    def __repr__(self):
        return "Link(%r, %r, %r)" % (self.target, self.label, self.link_target)

    def get_target(self):
        return self.target

    def get_label(self):
        return self.label or self.target

    def get_link_target(self):
        return self.link_target

class Linker:

    "Translate Moin links into other forms."

    # Pagename whitespace conversion, configured using the "whitespace" metadata
    # setting.

    default_whitespace_map = [
        (" ", "_"),
        ("\n", "_"),
        ]

    def __init__(self, metadata):

        "Initialise the linker with the 'metadata'."

        self.metadata = metadata

        # Obtain essential metadata.

        self.mapping = metadata.get("mapping", {})
        self.root_pagename = metadata.get("root_pagename")
        self.attachments_dir = metadata.get("attachments")

def resolve(path, pagename, root_pagename):

    "Resolve 'path' relative to 'pagename'."

    # Omit the root pagename from the resolved path components.

    if pagename == root_pagename:
        parts = []
    else:
        parts = pagename.rstrip("/").split("/")

    t = path.split("/")

    first = True

    for p in t:

        # Handle replacement of the page with another.

        if p == ".":
            parts = []

        # Handle ascent in the page hierarchy.

        elif p == "..":
            if parts:
                parts.pop()

        # Any non-navigation element replaces the path at the start.
        # Otherwise, the path is extended.
        # Omit the root pagename from the resolved path components if it would
        # appear at the start.

        elif p:
            if first:
                if p == root_pagename:
                    parts = []
                else:
                    parts = [p]
            else:
                if parts or p != root_pagename:
                    parts.append(p)

        first = False

    return "/".join(parts)

# vim: tabstop=4 expandtab shiftwidth=4
