#!/usr/bin/env python

"""
Output context common functionality.

Copyright (C) 2018 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import codecs

class Output:

    "A common output context abstraction."

    default_encoding = "utf-8"

    def __init__(self, metadata):

        "Initialise the output context with the 'metadata'."

        self.metadata = metadata

        # Obtain essential metadata.

        self.encoding = metadata.get("output_encoding", self.default_encoding)
        self.reset()

    def reset(self):

        "Set up an output collector."

        self.output = []

    def encode(self, text):

        "Encode 'text' using the configured encoding."

        return encode(text, self.encoding)

    def out(self, text):

        "Add 'text' to the output collector."

        self.output.append(text)

    # Page characteristics.

    def parent(self, pagename):

        "Return the parent of 'pagename'."

        return "/" in pagename and pagename.rsplit("/", 1)[0] or None

    # Serialisation methods.

    def to_string(self):

        "Return the output as a single Unicode string."

        s = u"".join(self.output)
        self.reset()
        return s

    def can_write(self):

        "Return whether this context supports page writing."

        return False

    def writefile(self, text, filename, encoding=None):

        """
        Write 'text' to the file having the given 'filename'. If the
        optional 'encoding' is specified, override the general encoding.

        Subclasses need to override this method for it to have an effect.
        """

        pass

    def writepage(self, text, pagename, encoding=None):

        """
        Write 'text' to the file having the given 'pagename' and optional
        'encoding'.
        """

        return self.writefile(text, self.to_filename(pagename), encoding)

    # Output methods.

    def writepath(self, text, filename, encoding=None):

        """
        Write 'text' to the file having the given 'filename'. If the
        optional 'encoding' is specified, override the general encoding.
        """

        f = codecs.open(filename, "w", encoding=encoding or self.encoding)
        try:
            f.write(text)
        finally:
            f.close()

def encode(s, encoding):

    "Encode 's' using 'encoding' if Unicode."

    if isinstance(s, unicode):
        return s.encode(encoding)
    else:
        return s

# vim: tabstop=4 expandtab shiftwidth=4
