#!/usr/bin/env python

"""
Moin wiki format parser.

Copyright (C) 2017, 2018, 2019 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

# Document transformations.

from moinformat.macros import get_macro

# Parser functionality and pattern definition.

from moinformat.parsers.common import ParserBase, get_patterns, choice, \
                                      excl, expect, group, optional, recur, \
                                      repeat

# Serialisation.

from moinformat.serialisers import serialise

# Document tree nodes.

from moinformat.tree.moin import Anchor, Break, Comment, DefItem, DefTerm, \
                                 Directive, FontStyle, Heading, Larger, \
                                 LineBreak, Link, LinkLabel, LinkParameter, \
                                 List, ListItem, Macro, Monospace, Region, \
                                 Rule, Smaller, Strikethrough, Subscript, \
                                 Superscript, Table, TableAttr, TableAttrs, \
                                 TableCell, TableRow, Text, Transclusion, \
                                 Underline, Verbatim

# Link parsing.

from moinformat.utils.links import parse_link_target

join = "".join

class MoinParser(ParserBase):

    "A wiki region parser."

    format = "moin"

    def __init__(self, metadata, parsers=None, root=None):

        """
        Initialise the parser with the given 'metadata' and optional 'parsers'.
        An optional 'root' indicates the document-level parser.
        """

        ParserBase.__init__(self, metadata, parsers, root)

        # Record certain node occurrences for later evaluation.

        self.macros = []

        # Record headings for identifier disambiguation.

        self.headings = []

        # Record link targets for resource identification.

        self.link_targets = []

    # Principal parser methods.

    def parse(self, s):

        """
        Parse page text 's'. Pages consist of regions delimited by markers.
        """

        self.items = self.get_items(s)
        self.region = Region([], type="moin")

        # Parse page header and directives.

        self.parse_region_header(self.region)
        self.parse_region_directives(self.region)

        # Handle pages directly with this parser. Pages do not need to use an
        # explicit format indicator.

        if not self.region.type:
            self.parse_region_content(self.items, self.region)

        # Otherwise, test the type and find an appropriate parser.

        else:
            self.parse_region_type(self.region)

        # Assign heading identifiers.

        self.identify_headings()

        return self.region



    # Macro evaluation.

    def evaluate_macros(self):

        "Evaluate the macro nodes in the document."

        for node in self.macros:

            # Obtain a class for the named macro.

            macro_cls = get_macro(node.name)
            if not macro_cls:
                continue

            # Instantiate the class and evaluate the macro.

            macro = macro_cls(node, self.region, self.metadata)
            macro.evaluate()

    # Metadata extraction.

    def update_metadata(self, metadata):

        "Update 'metadata' for the document."

        if self.headings:
            metadata.set("title", self.headings[0].text_content())
        else:
            metadata.set("title", self.metadata.get("pagename"))

    # Heading disambiguation.

    def identify_headings(self):

        "Assign identifiers to headings based on their textual content."

        d = {}

        for heading in self.headings:
            text = heading.text_content()

            if not d.has_key(text):
                d[text] = 0
                heading.identifier = text
            else:
                d[text] += 1
                heading.identifier = "%s-%d" % (text, d[text])



    # Conversion back to text.

    def get_serialiser(self):

        "Return metadata employing Moin as the output format."

        metadata = self.metadata.copy()
        metadata.set("link_format", None)
        metadata.set("output_context", "standalone")
        metadata.set("output_format", "moin")
        return metadata.get_serialiser()



    # Parser methods supporting different page features.

    def parse_attrname(self, attrs):

        "Handle an attribute name within 'attrs'."

        name = self.match_group("name")
        attr = TableAttr(name)

        preceding = self.read_until(["attrvalue"], False)
        if preceding == "":
            attr.quote = self.match_group("quote")
            attr.value = self.match_group("value")

        attrs.append(attr)

    def parse_break(self, region):

        "Handle a paragraph break within 'region'."

        self.add_node(region, Break())
        self.new_block(region)

    def parse_comment(self, region):

        "Handle a comment within 'region'."

        comment = self.match_group("comment")
        extra = self.match_group("extra")
        self.add_node(region, Comment(comment, extra))
        self.new_block(region)

    def parse_defitem(self, region, extra=""):

        "Handle a definition item within 'region'."

        pad = self.match_group("pad")
        item = DefItem([], pad, extra)
        self.parse_region_details(item, self.listitem_pattern_names)
        self.add_node(region, item)
        self.new_block(region)

    def parse_defterm(self, region):

        "Handle a definition term within 'region'."

        pad = self.match_group("pad")
        term = DefTerm([], pad)
        self.parse_region_details(term, ["deftermend", "deftermsep"] + self.inline_pattern_names)
        self.add_node(region, term)

        if self.matching_pattern() == "deftermsep":
            self.parse_defitem(region)

        # Add padding from the separator to the term, there being no item.

        else:
            term.extra = self.match_group("pad")

    def parse_defterm_empty(self, region):

        "Handle an empty definition term within 'region'."

        extra = self.match_group("pad")
        self.parse_region_details(region, ["deftermsep"])
        self.parse_defitem(region, extra)

    def parse_directive(self, region):

        "Handle a processing directive within 'region'."

        directive = self.match_group("directive")
        extra = self.match_group("extra")
        self.add_node(region, Directive(directive, extra))
        self.new_block(region)

    def parse_fontstyle(self, region):

        "Handle emphasis and strong styles."

        n = len(self.match_group("style"))

        # Handle endings.

        if isinstance(region, FontStyle):
            emphasis = n in (2, 4, 5)
            strong = n in (3, 5, 6)
            active = True

            if region.emphasis and emphasis:
                active = region.close_emphasis()
                n -= 2
            if region.strong and strong:
                active = region.close_strong()
                n -= 3

            if not active:
                if n:
                    self.items.rewind(n)
                raise StopIteration

            elif not n:
                return

        # Handle new styles.

        emphasis = n in (2, 4, 5)
        strong = n in (3, 5, 6)
        double = n in (4, 6)

        span = FontStyle([], emphasis, strong)
        if not double:
            self.parse_region_details(span, self.inline_pattern_names)
        region.append_inline(span)

    def parse_halign(self, attrs):

        "Handle horizontal alignment within 'attrs'."

        value = self.match_group("value")
        attr = TableAttr("align", value == "(" and "left" or value == ")" and "right" or "center", True)
        attrs.append(attr)

    def parse_heading(self, region):

        "Handle a heading."

        start_extra = self.match_group("extra")
        level = len(self.match_group("level"))
        start_pad = self.match_group("pad")
        heading = Heading([], level, start_extra, start_pad)
        self.parse_region_details(heading, ["headingend"] + self.inline_pattern_names)
        self.add_node(region, heading)
        self.new_block(region)

        # Record the heading for later processing.

        self.root.headings.append(heading)

    def parse_heading_end(self, heading):

        "Handle the end of a heading."

        level = len(self.match_group("level"))
        if heading.level == level:
            heading.end_pad = self.match_group("pad")
            heading.end_extra = self.match_group("extra")
            raise StopIteration

    def parse_list(self, item):

        "Create a list, starting with 'item'."

        list = List([item])
        self.parse_region_details(list, self.list_pattern_names, True)
        return list

    def parse_listitem(self, region):

        "Handle a list item marker within 'region'."

        indent = len(self.match_group("indent"))
        marker = self.match_group("marker")
        num = self.match_group("num")
        space = self.match_group("pad")

        last = region.node(-1)

        new_list = not isinstance(last, (List, ListItem))
        same_indent = not new_list and indent == last.indent
        new_marker = not new_list and last.marker != marker and same_indent
        new_num = not new_list and num is not None and last.num != num and same_indent

        # If the marker or number changes at the same indent, or if the indent
        # is smaller, queue the item and end the list.

        # Note that Moin format does not seek to support item renumbering,
        # instead starting new lists on number changes.

        if not new_list and (new_marker or new_num or indent < last.indent):
            self.queue_match()
            self.end_region(region)

        # Obtain a list item and populate it.

        item = ListItem([], indent, marker, space, num)
        self.parse_region_details(item, self.listitem_pattern_names)

        # Start a new list if not preceded by a list item, adding a trailing
        # block for new elements.

        if new_list:
            item = self.parse_list(item)
            self.add_node(region, item)
            self.new_block(region)

        # Add a nested list to the last item.

        elif indent > last.indent:
            item = self.parse_list(item)
            self.add_node(last, item)

        # Add the item to the current list.

        else:
            self.add_node(region, item)

    def parse_rule(self, region):

        "Handle a horizontal rule within 'region'."

        height = len(self.match_group("rule")) - 4
        rule = Rule(height)
        self.add_node(region, rule)
        self.new_block(region)

    def parse_section(self, region):

        "Handle the start of a new section within 'region'."

        # Parse the section and start a new block after the section.

        indent = len(self.match_group("indent"))
        level = len(self.match_group("level"))

        section = self.parse_region(level, indent, "inline")

        # If the section is inline, treat it like any other inline element.

        if section.type == "inline":
            region.append_inline(section)

        # Otherwise, add it as a new block element.

        else:
            self.add_node(region, section)
            if region.allow_blocks:
                self.new_block(region)

    def parse_table_attrs(self, cell):

        "Handle the start of table attributes within 'cell'."

        attrs = TableAttrs([])
        self.parse_region_details(attrs, self.table_attr_pattern_names)

        # Test the validity of the attributes.

        last = None

        for node in attrs.nodes:

            # Text separator nodes must be whitespace.

            if isinstance(node, Text):
                if node.s.strip():
                    break

            # Named attributes must be preceded by space if not the first.

            elif last and not node.concise and not isinstance(last, Text):
                break

            last = node

        # All nodes were valid: preserve the collection.

        else:
            # Add the attributes as a node, also recording their presence.

            cell.append(attrs)
            cell.attrs = attrs
            return

        # Invalid nodes were found: serialise the attributes as text.

        cell.append_inline(Text(serialise(attrs, self.get_serialiser())))

    def parse_table_row(self, region):

        "Handle the start of a table row within 'region'."

        # Identify any active table.

        table = region.node(-2)
        block = region.node(-1)

        if not (isinstance(table, Table) and block.empty()):
            new_table = table = Table([])
        else:
            new_table = None

        row = TableRow([])

        while True:
            cell = TableCell([])
            self.parse_region_details(cell, self.table_row_pattern_names)

            # Handle the end of the row.

            if self.matching_pattern() == "tableend":
                trailing = self.match_group("extra")

                # If the cell was started but not finished, convert the row into text.

                if not row.nodes or not cell.empty():

                    # Convert the nodes back to text.

                    serialiser = self.get_serialiser()

                    for node in row.nodes:
                        region.append_inline(Text(serialise(node, serialiser)))

                    region.append_inline(Text(serialise(cell, serialiser) + trailing))

                    self.new_block(region)
                    return

                # Append the final cell, if not empty.

                else:
                    row.trailing = trailing

                    if not cell.empty():
                        row.append(cell)
                    break

            # A cell separator has been found.

            row.append(cell)

        # Add the row to the table and any new table to the region.

        table.add(row)
        if new_table:
            self.add_node(region, new_table)

        self.new_block(region)

    def parse_valign(self, attrs):

        "Handle vertical alignment within 'attrs'."

        value = self.match_group("value")
        attr = TableAttr("valign", value == "^" and "top" or "bottom", True)
        attrs.append(attr)



    def inline_patterns_for(self, name):

        "Return active patterns for the inline element having the given 'name'."

        names = self.inline_pattern_names[:]
        names[names.index(name)] = "%send" % name
        return names



    # Inline formatting handlers.

    def parse_inline(self, region, cls, pattern_name):

        "Handle an inline region."

        span = cls([])
        self.parse_region_details(span, self.inline_patterns_for(pattern_name))
        region.append_inline(span)

    def parse_larger(self, region):
        self.parse_inline(region, Larger, "larger")

    def parse_monospace(self, region):
        span = Monospace([])
        self.parse_region_details(span, ["monospaceend"])
        region.append_inline(span)

    def parse_smaller(self, region):
        self.parse_inline(region, Smaller, "smaller")

    def parse_strike(self, region):
        self.parse_inline(region, Strikethrough, "strike")

    def parse_sub(self, region):
        self.parse_inline(region, Subscript, "sub")

    def parse_super(self, region):
        self.parse_inline(region, Superscript, "super")

    def parse_underline(self, region):
        self.parse_inline(region, Underline, "underline")

    # Link formatting handlers.

    def _parse_link(self, region, cls, pattern_names):
        target = self.match_group("target")
        end = self.match_group("end")

        # Obtain an object for the link target.

        link_target = parse_link_target(target, self.metadata)

        # Obtain an object for the node.

        span = cls([], link_target)

        # Obtain the extra details.

        if not end:
            cls = LinkLabel

            # Introduce a label or parameter for each separated region.

            while True:
                param = cls([])
                self.parse_region_details(param, pattern_names)
                span.append(param)

                if self.matching_pattern() != "linksep":
                    break

                cls = LinkParameter

        region.append_inline(span)

        # Record the link target for later processing.

        self.root.link_targets.append(link_target)

    def parse_link(self, region):
        self._parse_link(region, Link, self.link_pattern_names)

    def parse_transclusion(self, region):
        self._parse_link(region, Transclusion, self.transclusion_pattern_names)



    # Complete inline pattern handlers.

    def parse_anchor(self, region):
        target = self.match_group("target")
        anchor = Anchor(target)
        region.append_inline(anchor)

    def parse_linebreak(self, region):
        region.append_inline(LineBreak())

    def parse_macro(self, region):
        name = self.match_group("name")
        args = self.match_group("args")

        # Obtain the raw arguments. Moin usually leaves it to the macro to
        # interpret the individual arguments.

        arglist = args and args.split(",") or []
        macro = Macro(name, arglist, region.append_point(), region)
        region.append_inline(macro)

        # Record the macro for later processing.

        self.root.macros.append(macro)

    def parse_verbatim(self, region):
        text = self.match_group("verbatim")
        region.append_inline(Verbatim(text))



    # Table attribute handlers.

    def parse_table_attr(self, attrs, pattern_name):

        "Handle a table attribute."

        attrs.append(TableAttr(pattern_name, self.match_group("value"), True))

    def parse_colour(self, cell):
        self.parse_table_attr(cell, "bgcolor")

    def parse_colspan(self, cell):
        self.parse_table_attr(cell, "colspan")

    def parse_rowspan(self, cell):
        self.parse_table_attr(cell, "rowspan")

    def parse_width(self, cell):
        self.parse_table_attr(cell, "width")



    # Regular expressions.

    syntax = {
        # Page regions:

        "regionstart"   : join((group("indent", r"\N*"),                        # ws... (optional)
                                group("level", repeat("[{]", 3)))),             # {{{...

        "regionend"     : join((r"\N*",                                         # ws... (optional)
                                group("feature", join((
                                    group("level", repeat("[}]", 3)),           # }}}...
                                    optional(group("extra", r"\n"))))))),       # nl (optional)

        # Region header and directives:

        "header"        : join(("#!",                                           # #!
                                group("args", ".*?"), "\n")),                   # text-excl-nl

        "directive"     : join((r"^#",                                          # #
                                group("directive", r".*?$"),                    # rest of line
                                optional(group("extra", r"\n")))),              # nl (optional)

        # Region contents:

        # Line-oriented patterns support features which require their own
        # separate lines.

        "break"         : r"^(\s*?)\n",                                         # blank line

        "comment"       : join((r"^##",                                         # ##
                                group("comment", r".*?$"),                      # rest of line
                                optional(group("extra", r"\n")))),              # nl (optional)

        "defterm"       : join(("^",
                                group("pad", r"\N+"),                           # ws...
                                expect(".+?::"))),                              # text ::

        "defterm_empty" : join(("^",
                                group("pad", r"\N+"),                           # ws...
                                expect("::\s+"))),                              # :: ws...

        "heading"       : join(("^",
                                group("extra", r"\N*"),                         # ws... (optional)
                                group("level", "=+"),                           # =...
                                group("pad", r"\s+"),                           # ws...
                                expect(join((r".*?\N+",                         # text
                                             recur("level"),                    # =...
                                             r"\N*$"))))),                      # ws... (optional)

        "listitem"      : join(("^",
                                group("indent", r"\N+"),                        # ws...
                                group("marker", r"\*"),                         # list-marker
                                group("pad", r"\s*"))),                         # ws... (optional)

        "listitem_num"  : join(("^",
                                group("indent", r"\N+"),                        # ws...
                                group("marker", r"\d+\."),                      # decimal-marker
                                optional(join(("#", group("num", r"\d+")))),    # # num (optional)
                                group("pad", r"\s+"))),                         # ws...

        "listitem_alpha": join(("^",
                                group("indent", r"\N+"),                        # ws...
                                group("marker", r"[aA]\."),                     # alpha-marker
                                optional(join(("#", group("num", r"\d+")))),    # # num (optional)
                                group("pad", r"\s+"))),                         # ws...

        "listitem_roman": join(("^",
                                group("indent", r"\N+"),                        # ws...
                                group("marker", r"[iI]\."),                     # roman-marker
                                optional(join(("#", group("num", r"\d+")))),    # # num (optional)
                                group("pad", r"\s+"))),                         # ws...

        "listitem_dot"  : join(("^",
                                group("indent", r"\N+"),                        # ws...
                                group("marker", r"\."),                         # dot-marker
                                group("pad", r"\s*"))),                         # ws... (optional)

        "tablerow"      : r"^\|\|",                                             # ||

        # Region contents:

        # Inline patterns are for markup features that appear within blocks.
        # The patterns below start inline spans that can contain other markup
        # features.

        "fontstyle"     : group("style", repeat("'", 2, 6)),                    # ''...

        # Trivial markup balancing is done below using the end features.

        "larger"        : join((r"~\+",                                         # ~+
                                expect(r"\P*?\+~"))),                           # ... +~

        "monospace"     : join((r"`",                                           # `
                                expect(r"\P*?`"))),                             # ... `

        "smaller"       : join((r"~-",                                          # ~-
                                expect(r"\P*?-~"))),                            # ... -~

        "strike"        : join((r"--\(",                                        # --(
                                expect(r"\P*?\)--"))),                          # ... )--

        "sub"           : join((r",,",                                          # ,,
                                expect(r"\P*?,,"))),                            # ... ,,

        "super"         : join((r"\^",                                          # ^
                                expect(r"\P*?\^"))),                            # ... ^

        "underline"     : join((r"__",                                          # __
                                expect(r"\P*?__"))),                            # ... __

        # Rules are treated as inline but, unlike the above, appear without
        # contents.

        "rule"          : group("rule", "-----*"),                              # ----...

        # Links and transclusions may start inline spans.

        "link"          : join((r"\[\[",                                        # [[
                                group("target", r"\P*?"),                       # ...
                                choice((r"\|",                                  # |
                                        group("end", r"]]"))))),                # ]]

        "transclusion"  : join((r"\{\{",                                        # {{
                                excl(r"\{"),                                    # not-{
                                group("target", r"\P*?"),                       # ...
                                choice((r"\|",                                  # |
                                        group("end", r"}}"))))),                # }}

        # Complete inline patterns are for markup features that do not support
        # arbitrary content within them:

        "anchor"        : join((r"\(\(",                                        # ((
                                group("target", ".*?"),                         # target
                                r"\)\)")),                                      # ))

        "linebreak"     : r"\\\\",                                              # \\

        "macro"         : join(("<<",                                           # <<
                                group("name", "\w+?"),                          # digit-letter...
                                optional(join((r"\(",                           # ( (optional)
                                               group("args", ".*?"),            # not-)...
                                               r"\)"))),                        # ) (optional)
                                ">>")),                                         # >>

        "verbatim"      : join(("<<<",                                          # <<<
                                group("verbatim", r"\P*?"),                     # ...
                                ">>>")),

        # Ending patterns for inline features:

        "largerend"     : r"\+~",                                               # +~
        "linkend"       : r"]]",                                                # ]]
        "monospaceend"  : r"`",                                                 # `
        "smallerend"    : r"-~",                                                # -~
        "strikeend"     : r"\)--",                                              # )--
        "subend"        : r",,",                                                # ,,
        "superend"      : r"\^",                                                # ^
        "transclusionend": r"}}",                                               # }}
        "underlineend"  : r"__",                                                # __

        # Heading contents:

        "headingend"    : join((group("pad", r"\N+"),                           # ws...
                                group("level", "=+"),                           # =...
                                group("extra", r"\N*\n"))),                     # ws (optional) nl

        # Link/transclusion contents:

        "linksep"       : r"\|",                                                # |

        # List contents:

        "deftermend"    : join(("::", group("pad", r"\s*?\n"))),                # ::
                                                                                # ws... (optional)
                                                                                # nl

        "deftermsep"    : join(("::", group("pad", r"\s+"))),                   # ::
                                                                                # ws...

        "listitemend"   : join((r"^",                                           # next line
                                choice((expect(r"[^\s]"),                       # without indent
                                        expect(r"\Z"),                          # end of string
                                        expect(r"\N+\*"),                       # or with ws... list-marker
                                        expect(r"\N+\d\."),                     # or with ws... decimal-marker
                                        expect(r"\N+[aA]\."),                   # or with ws... alpha-marker
                                        expect(r"\N+[iI]\."),                   # or with ws... roman-marker
                                        expect(r"\N+\."),                       # or with ws... dot-marker
                                        expect(r"\N+.+?::\s"),                  # or with ws... text :: ws (next defterm)
                                        expect(r"\N+::\s"))))),                 # or with ws... :: ws (next defitem)

        # Table contents:

        "tableattrs"    : join(("<",                                            # lt
                                excl("<"))),                                    # not-lt

        "tablecell"     : r"\|\|",                                              # ||

        "tableend"      : join((group("extra", r"\s*?"),                        # ws... (optional)
                                "^")),                                          # next line

        # Table attributes:

        "tableattrsend" : r">",                                                 # >
        "halign"        : group("value", "[(:)]"),                              # halign-marker
        "valign"        : group("value", "[v^]"),                               # valign-marker
        "colour"        : group("value", join(("\#",                            # #
                                               repeat("[0-9A-Fa-f]", 6, 6)))),  # nnnnnn

        "colspan"       : join(("-",                                            # -
                                group("value", "\d+"))),                        # n...

        "rowspan"       : join((r"\|",                                          # |
                                group("value", "\d+"))),                        # n...

        "width"         : group("value", "\d+%"),                               # n... %

        "attrname"      : join((excl(r"[-\d]"),                                 # not-dash-or-digit
                                group("name", r"[-\w]+"))),                     # dash-digit-letter...

        "attrvalue"     : join(("=", group("quote", r"\Q"),                     # quote
                                     group("value", ".*?"),                     # non-quote... (optional)
                                     recur("quote"))),                          # quote
        }

    patterns = get_patterns(syntax)



    # Patterns available within certain markup features.

    table_attr_pattern_names = [
        "attrname", "colour", "colspan", "halign", "rowspan", "tableattrsend",
        "valign", "width"
        ]

    inline_without_links_pattern_names = [
        "anchor", "fontstyle", "larger", "linebreak", "macro",
        "monospace", "regionstart", "smaller", "strike", "sub", "super",
        "underline", "verbatim"
        ]

    inline_pattern_names = inline_without_links_pattern_names + [
        "link", "transclusion"]

    link_pattern_names = inline_without_links_pattern_names + [
        "linkend", "linksep", "transclusion"]

    list_pattern_names = [
        "listitem", "listitem_alpha", "listitem_dot", "listitem_num",
        "listitem_roman",
        ]

    listitem_pattern_names = inline_pattern_names + ["listitemend"]

    region_without_table_pattern_names = inline_pattern_names + list_pattern_names + [
        "break", "comment", "heading", "defterm", "defterm_empty",
        "regionend", "rule",
        ]

    table_row_pattern_names = inline_pattern_names + [
        "tableattrs", "tablecell", "tableend"
        ]

    transclusion_pattern_names = inline_without_links_pattern_names + [
        "linksep", "transclusionend"]

    # The region pattern names are specifically used by the common parser
    # functionality.

    region_pattern_names = region_without_table_pattern_names + ["tablerow"]



    # Pattern handlers.

    end_region = ParserBase.end_region
    parse_section_end = ParserBase.parse_region_end

    handlers = {
        None : end_region,
        "anchor" : parse_anchor,
        "attrname" : parse_attrname,
        "break" : parse_break,
        "colour" : parse_colour,
        "colspan" : parse_colspan,
        "comment" : parse_comment,
        "defterm" : parse_defterm,
        "defterm_empty" : parse_defterm_empty,
        "deftermend" : end_region,
        "deftermsep" : end_region,
        "directive" : parse_directive,
        "fontstyle" : parse_fontstyle,
        "halign" : parse_halign,
        "heading" : parse_heading,
        "headingend" : parse_heading_end,
        "larger" : parse_larger,
        "largerend" : end_region,
        "linebreak" : parse_linebreak,
        "link" : parse_link,
        "linkend" : end_region,
        "linksep" : end_region,
        "macro" : parse_macro,
        "listitemend" : end_region,
        "listitem" : parse_listitem,
        "listitem_alpha" : parse_listitem,
        "listitem_dot" : parse_listitem,
        "listitem_num" : parse_listitem,
        "listitem_roman" : parse_listitem,
        "monospace" : parse_monospace,
        "monospaceend" : end_region,
        "regionstart" : parse_section,
        "regionend" : parse_section_end,
        "rowspan" : parse_rowspan,
        "rule" : parse_rule,
        "smaller" : parse_smaller,
        "smallerend" : end_region,
        "strike" : parse_strike,
        "strikeend" : end_region,
        "sub" : parse_sub,
        "subend" : end_region,
        "super" : parse_super,
        "superend" : end_region,
        "tableattrs" : parse_table_attrs,
        "tableattrsend" : end_region,
        "tablerow" : parse_table_row,
        "tablecell" : end_region,
        "tableend" : end_region,
        "transclusion" : parse_transclusion,
        "transclusionend" : end_region,
        "underline" : parse_underline,
        "underlineend" : end_region,
        "valign" : parse_valign,
        "verbatim" : parse_verbatim,
        "width" : parse_width,
        }

parser = MoinParser

# vim: tabstop=4 expandtab shiftwidth=4
