#!/usr/bin/env python

"""
Moin wiki table parser.

Copyright (C) 2017, 2018 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from moinformat.parsers.common import get_patterns, \
                                      excl, expect, group
from moinformat.parsers.moin import MoinParser
from moinformat.tree.moin import Table, TableAttrs, TableCell, TableRow, Text
from moinformat.tree.table import Continuation

join = "".join

# Parser functionality.

class TableParser(MoinParser):

    "A parser for improved table syntax."

    format = "table"

    # Principal parser methods.

    def parse_region_content(self, items, region):

        """
        Parse the data provided by 'items' to populate the given 'region'. For
        table regions, normal region handling is wrapped by management of the
        table structure.
        """

        self.set_region(items, region)

        # Start to populate table rows.

        cell = TableCell([])
        row = TableRow([cell])
        table = Table([row])
        self.append_node(self.region, table)

        while True:
            self.parse_region_details(cell, self.region_pattern_names)

            # Detect the end of the table.

            pattern = self.matching_pattern()

            if pattern == "regionend":
                break

            elif pattern == "columnsep":
                cell = TableCell([])
                row.append(cell)

            elif pattern == "rowsep":
                row = TableRow([])
                table.append(row)
                cell = TableCell([])
                row.append(cell)

    # Parser handler methods.

    def parse_continuation(self, cell):

        "Handle continuation padding."

        feature = self.match_group("feature")
        cell.append(Continuation(feature))



    # Regular expressions.

    syntax = {}
    syntax.update(MoinParser.syntax)
    syntax.update({
        # At start of line:

        "rowsep"        : join(("^==",                      # ==
                                excl(r".*==\s*?$"),         # not-heading
                                expect(r"\N*?"))),          # ws-excl-nl

        "continuation"  : group("feature",
                          join(("^",
                                group("indent", r"\N*"),    # ws... (optional)
                                r"\.\.",                    # ..
                                excl(r"\.")))),             # not-.

        # Within text:

        "columnsep"     : join((r"\|\|",                    # ||
                                excl(r"\|"))),              # not-|
        })

    patterns = get_patterns(syntax)



    # Pattern details.

    region_pattern_names = [
        "columnsep", "continuation", "rowsep", "tableattrs",
        ] + MoinParser.region_without_table_pattern_names



    # Pattern handlers.

    end_region = MoinParser.end_region
    parse_table_end = MoinParser.parse_region_end

    handlers = {}
    handlers.update(MoinParser.handlers)
    handlers.update({
        "columnsep" : end_region,
        "continuation" : parse_continuation,
        "rowsep" : end_region,
        "regionend" : parse_table_end,
        })

parser = TableParser

# vim: tabstop=4 expandtab shiftwidth=4
