#!/usr/bin/env python

"""
HTML serialiser.

Copyright (C) 2017, 2018, 2019 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from moinformat.serialisers.common import escape_attr, escape_text, Serialiser
from moinformat.tree.moin import LinkLabel, LinkParameter, Text
from moinformat.utils.links import parse_link_target

class HTMLSerialiser(Serialiser):

    "Serialisation of the page."

    format = "html"

    def _region_tag(self, type):

        # NOTE: Need to support types in general.

        type = type and type.split()[0]

        if type == "inline":
            return "tt"
        elif type in (None, "python"):
            return "pre"
        else:
            return "span"

    def start_region(self, level, indent, type, extra):

        # Generate attributes, joining them when preparing the tag.

        l = []
        out = l.append

        if level:
            out("level-%d" % level)

        if indent:
            out("indent-%d" % indent)

        # NOTE: Encode type details for CSS.

        out("type-%s" % escape_attr(type or "opaque"))

        tag = self._region_tag(type)

        # Inline regions must preserve "indent" as space in the text.

        if type == "inline" and indent:
            self.out(" " * indent)

        self.out("<%s class='%s'>" % (tag, " ".join(l)))

    def end_region(self, level, indent, type, extra):
        tag = self._region_tag(type)
        self.out("</%s>" % tag)

    def start_block(self):
        self.out("<p>")

    def end_block(self):
        self.out("</p>")

    def start_defitem(self, pad, extra):
        self.out("<dd>")

    def end_defitem(self, pad, extra):
        self.out("</dd>")

    def start_defterm(self, pad, extra):
        self.out("<dt>")

    def end_defterm(self, pad, extra):
        self.out("</dt>")

    def start_emphasis(self):
        self.out("<em>")

    def end_emphasis(self):
        self.out("</em>")

    def start_heading(self, level, extra, pad, identifier):
        self.out("<h%d id='%s'>" % (level, escape_attr(self.linker.make_id(identifier))))

    def end_heading(self, level, pad, extra):
        self.out("</h%d>" % level)

    def start_larger(self):
        self.out("<big>")

    def end_larger(self):
        self.out("</big>")

    def start_linktext(self):
        pass

    def end_linktext(self):
        pass

    list_tags = {
        "i" : "lower-roman",
        "I" : "upper-roman",
        "a" : "lower-latin",
        "A" : "upper-latin",
        }

    def _get_list_tag(self, marker):
        if marker:
            if marker[0].isdigit():
                return "ol", "decimal"
            style_type = self.list_tags.get(marker[0])
            if style_type:
                return "ol", style_type

        return "ul", None

    def start_list(self, indent, marker, num):
        tag, style_type = self._get_list_tag(marker)
        style = style_type and ' style="list-style-type: %s"' % escape_attr(style_type) or ""
        start = style_type and num is not None and ' start="%s"' % escape_attr(num) or ""
        self.out("<%s%s%s>" % (tag, style, start))

    def end_list(self, indent, marker, num):
        tag, style = self._get_list_tag(marker)
        self.out("</%s>" % tag)

    def start_listitem(self, indent, marker, space, num):
        self.out("<li>")

    def end_listitem(self, indent, marker, space, num):
        self.out("</li>")

    def start_macro(self, name, args, nodes, inline):
        tag = inline and "span" or "div"
        self.out("<%s class='macro %s'>" % (tag, escape_text(name)))

        # Fallback case for when macros are not replaced.

        if not nodes:
            self.out(escape_text("<<"))
            self.out("<span class='name'>%s</span>" % escape_text(name))
            if args:
                self.out("(")
            first = True
            for arg in args:
                if not first:
                    self.out(",")
                self.out("<span class='arg'>%s</span>" % escape_text(arg))
                first = False
            if args:
                self.out(")")
            self.out(escape_text(">>"))

    def end_macro(self, inline):
        tag = inline and "span" or "div"
        self.out("</%s>" % tag)

    def start_monospace(self):
        self.out("<tt>")

    def end_monospace(self):
        self.out("</tt>")

    def start_smaller(self):
        self.out("<small>")

    def end_smaller(self):
        self.out("</small>")

    def start_strikethrough(self):
        self.out("<del>")

    def end_strikethrough(self):
        self.out("</del>")

    def start_strong(self):
        self.out("<strong>")

    def end_strong(self):
        self.out("</strong>")

    def start_subscript(self):
        self.out("<sub>")

    def end_subscript(self):
        self.out("</sub>")

    def start_superscript(self):
        self.out("<sup>")

    def end_superscript(self):
        self.out("</sup>")

    def start_table(self):
        self.out("<table>")

    def end_table(self):
        self.out("</table>")

    def start_table_attrs(self):
        pass

    def end_table_attrs(self):
        pass

    def start_table_cell(self, attrs):
        self.out("<td")

        # Handle the attributes separately from their container.

        if attrs and not attrs.empty():
            for attr in attrs.nodes:
                attr.to_string(self)

        self.out(">")

    def end_table_cell(self):
        self.out("</td>")

    def start_table_row(self):
        self.out("<tr>")

    def end_table_row(self, trailing):
        self.out("</tr>")

    def start_underline(self):
        self.out("<span style='text-decoration: underline'>")

    def end_underline(self):
        self.out("</span>")

    def anchor(self, target):
        self.out("<a name='%s' />" % escape_attr(self.linker.make_id(target)))

    def break_(self):
        pass

    def comment(self, comment, extra):
        pass

    def directive(self, directive, extra):
        name, text = directive.split(None, 1)

        # Produce a readable redirect.

        if name.lower() == "redirect":
            self.start_block()

            # Process the redirect argument as a link target, producing a link
            # element.

            target = parse_link_target(text, self.metadata)
            self._link(target, [LinkLabel([Text(text)])], "a", "href")

            self.end_block()

    def linebreak(self):
        self.out("<br />")

    def _link(self, target, nodes, tag, attr):
        link = self.linker and self.linker.translate(target) or None

        self.out('<%s %s="%s"' % (tag, attr, escape_attr(link.get_target())))

        # Provide link parameters as attributes.

        if nodes:
            for node in nodes:
                if isinstance(node, LinkParameter):
                    self.out(" ")
                    node.to_string(self)

        # Close the tag if an image.

        if tag == "img":
            self.out(" />")

        # Provide the link label if specified. Otherwise, use a generated
        # default for the label.

        else:
            self.out(">")

            for node in nodes or []:
                if isinstance(node, LinkLabel):
                    node.to_string(self)
                break
            else:
                self.out(escape_text(link.get_label()))

            self.out("</%s>" % tag)

    def link(self, target, nodes):
        self._link(target, nodes, "a", "href")

    def link_label(self, nodes):
        for node in nodes:
            node.to_string(self)

    def link_parameter(self, key_value):
        if len(key_value) == 1:
            self.out(key_value[0])
        else:
            key, value = key_value
            self.out("%s='%s'" % (key, escape_attr(value)))

    def rule(self, height):
        self.out("<hr style='height: %dpt' />" % min(height, 10))

    def table_attrs(self, nodes):

        # Skip the attributes in their original form.

        pass

    def table_attr(self, name, value, concise, quote):
        self.out(" %s%s" % (escape_text(name), value is not None and
            "='%s'" % escape_attr(value) or ""))

    def text(self, s):
        self.out(escape_text(s))

    def transclusion(self, target, nodes):
        self._link(target, nodes, "img", "src")

    def verbatim(self, s):
        self.text(s)

serialiser = HTMLSerialiser

# vim: tabstop=4 expandtab shiftwidth=4
