#!/usr/bin/env python

"""
Common HTML theme functionality.

Copyright (C) 2018, 2019 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from moinformat.themes.common import Theme
from os import listdir
from os.path import splitext

class HTMLTheme(Theme):

    "A common HTML theme abstraction."

    # Support a collection of links to stylesheets provided by each theme.

    link = '<link rel="stylesheet" type="text/css" charset="utf-8"' \
           ' media="%(media)s" href="%(root)s/_css/%(filename)s" />\n'

    bundled_link = '<style type="text/css" media="%(media)s">\n%(data)s' \
                   '</style>\n'

    def get_links(self, subs, bundle=False):

        """
        Using 'subs', return a string containing markup linking to resources if
        'bundle' is a false value or omitted. If 'bundle' is a true value,
        incorporate resources within the document.
        """

        d = {}
        d.update(subs)

        css = self.get_resource("css")
        links = []

        for filename in listdir(css):

            # Only link to CSS files.

            basename, ext = splitext(filename)
            if ext != ".css":
                continue

            # Filenames can have the form <media>.css or <media>-<name>.css to
            # set the media type.

            t = basename.split("-", 1)

            d["media"] = t[0] or "all"
            d["filename"] = filename

            if bundle:
                d["data"] = self.load_resource(filename, css)

            template = bundle and self.bundled_link or self.link
            links.append(template % d)

        return "".join(links)

    # Public methods.

    def apply(self, text):

        "Apply this theme to the given 'text', returning a themed version."

        template = self.load_resource("template.html")
        subs = {
            "encoding" : self.output.encoding,
            "root" : self.linker.get_top_level() or ".",
            "text" : text,
            "title" : self.metadata.get("title") or self.metadata.get("pagename"),
            }
        subs["links"] = self.get_links(subs, self.bundle)
        return template % subs

    def install_resources(self):

        "Install resources for this theme."

        if not self.bundle:
            self.install_resource("css", "_css")

# vim: tabstop=4 expandtab shiftwidth=4
