#!/usr/bin/env python

"""
Moin wiki format document tree nodes.

Copyright (C) 2017, 2018, 2019 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

class Container:

    "A container of document nodes."

    def __init__(self, nodes):
        self.nodes = nodes

        # In principle, allow blocks within containers. Some nodes may forbid
        # them to simplify the document structure.

        self.allow_blocks = True

    def append(self, node):
        self.nodes.append(node)

    def append_many(self, nodes):
        for node in nodes:
            self.append(node)

    add = append

    def append_inline(self, node):

        "Append 'node' inline within the appropriate container."

        n = self.append_point()

        # Redirect the addition if another container is to accept the node.

        if n is not self:
            n.append_inline(node)

        # Otherwise, append within this container.

        else:
            n.append(node)

    def append_inline_many(self, nodes):
        for node in nodes:
            self.append_inline(node)

    def append_point(self):

        "Return the container to which inline nodes are added."

        last = self.node(-1)

        if isinstance(last, Block):
            return last
        else:
            return self

    def empty(self):
        return not self.nodes

    def insert_after(self, old, new):

        "Insert after 'old' in the children the 'new' node."

        index = self.nodes.index(old)
        self.nodes.insert(index + 1, new)

    def node(self, index):
        try:
            return self.nodes[index]
        except IndexError:
            return None

    def normalise(self):

        "Combine adjacent text nodes."

        nodes = self.nodes
        self.nodes = []
        text = None

        for node in nodes:

            # Open a text node or merge text into an open node.

            if isinstance(node, Text):
                if not text:
                    text = node
                else:
                    text.merge(node)

            # Close any open text node and append the current node.

            else:
                if text:
                    self.append(text)
                    text = None
                self.append(node)

        # Add any open text node.

        if text:
            self.append(text)

    def remove(self, node):

        "Remove 'node' from the children."

        self.nodes.remove(node)

    def replace(self, old, new):

        "Replace 'old' with 'new' in the children."

        i = self.nodes.index(old)
        self.nodes[i] = new

    def split_at(self, node):

        """
        Split the container at 'node', returning a new container holding the
        nodes following 'node' that are moved from this container.
        """

        i = self.nodes.index(node)
        following = self.__class__(self.nodes[i+1:])

        # Remove the node and the following parts from this container.

        del self.nodes[i:]
        return following

    def text_content(self):

        """
        Return a string containing the content of text nodes within this
        container.
        """

        l = []

        for node in self.nodes:
            if isinstance(node, Text):
                l.append(node.s)
            elif isinstance(node, Container):
                l.append(node.text_content())

        return "".join(l)

    def whitespace_only(self):

        "Return whether the container provides only whitespace text."

        return not self.text_content().strip()

    def __str__(self):
        return self.prettyprint()

    def _prettyprint(self, l, indent=""):
        for node in self.nodes:
            l.append(node.prettyprint(indent + "  "))
        return "\n".join(l)

    def _to_string(self, out):
        for node in self.nodes:
            node.to_string(out)

class Region(Container):

    "A region of the page."

    def __init__(self, nodes, level=0, indent=0, type=None, args=None,
                 transparent=True, extra=None):
        Container.__init__(self, nodes)
        self.level = level
        self.indent = indent
        self.type = type
        self.args = args
        self.transparent = transparent
        self.extra = extra

    def add(self, node):
        last = self.node(-1)
        if last and last.empty():
            self.nodes[-1] = node
        else:
            self.append(node)

    def append_point(self):

        "Return the container to which inline nodes are added."

        if self.transparent:
            return self.nodes[-1]
        else:
            return self

    def have_end(self, s):
        return self.level and s.startswith("}") and self.level == len(s)

    def __repr__(self):
        return "Region(%r, %r, %r, %r, %r, %r, %r)" % (self.nodes, self.level,
            self.indent, self.type, self.args, self.transparent, self.extra)

    def prettyprint(self, indent=""):
        l = ["%sRegion: level=%d indent=%d type=%s args=%r extra=%r" % (indent,
             self.level, self.indent, self.type, self.args, self.extra)]
        return self._prettyprint(l, indent)

    def to_string(self, out):
        out.start_region(self.level, self.indent, self.type, self.extra)

        # Obtain a serialiser for the region from the same format family.
        # Retain the same serialiser if no appropriate serialiser could be
        # obtained.

        serialiser_name = "%s.%s" % (out.format, self.type)
        serialiser = out.get_serialiser(serialiser_name)

        # Serialise the region.

        self._to_string(serialiser)

        # End the region with the previous serialiser.

        out.end_region(self.level, self.indent, self.type, self.extra)



# Block nodes.

class Block(Container):

    "A block in the page."

    def __repr__(self):
        return "Block(%r)" % self.nodes

    def prettyprint(self, indent=""):
        l = ["%sBlock" % indent]
        return self._prettyprint(l, indent)

    def to_string(self, out):
        out.start_block()
        self._to_string(out)
        out.end_block()

class DefItem(Container):

    "A definition item."

    def __init__(self, nodes, pad, extra):
        Container.__init__(self, nodes)
        self.pad = pad
        self.extra = extra

    def __repr__(self):
        return "DefItem(%r, %r, %r)" % (self.nodes, self.pad, self.extra)

    def prettyprint(self, indent=""):
        l = ["%sDefItem: pad=%r extra=%r" % (indent, self.pad, self.extra)]
        return self._prettyprint(l, indent)

    def to_string(self, out):
        out.start_defitem(self.pad, self.extra)
        self._to_string(out)
        out.end_defitem(self.pad, self.extra)

class DefTerm(Container):

    "A definition term."

    def __init__(self, nodes, pad, extra=""):
        Container.__init__(self, nodes)
        self.pad = pad
        self.extra = extra

    def __repr__(self):
        return "DefTerm(%r, %r, %r)" % (self.nodes, self.pad, self.extra)

    def prettyprint(self, indent=""):
        l = ["%sDefTerm: pad=%r extra=%r" % (indent, self.pad, self.extra)]
        return self._prettyprint(l, indent)

    def to_string(self, out):
        out.start_defterm(self.pad, self.extra)
        self._to_string(out)
        out.end_defterm(self.pad, self.extra)

class FontStyle(Container):

    "Emphasised and/or strong text."

    def __init__(self, nodes, emphasis=False, strong=False):
        Container.__init__(self, nodes)
        self.emphasis = emphasis
        self.strong = strong

    def close_emphasis(self):
        if self.strong:
            span = FontStyle(self.nodes, emphasis=True)
            self.nodes = [span]
            self.emphasis = False
        return self.strong

    def close_strong(self):
        if self.emphasis:
            span = FontStyle(self.nodes, strong=True)
            self.nodes = [span]
            self.strong = False
        return self.emphasis

    def __repr__(self):
        return "FontStyle(%r, %r, %r)" % (self.nodes, self.emphasis, self.strong)

    def prettyprint(self, indent=""):
        l = ["%sFontStyle: emphasis=%r strong=%r" % (indent, self.emphasis, self.strong)]
        return self._prettyprint(l, indent)

    def to_string(self, out):
        if self.emphasis:
            out.start_emphasis()
        elif self.strong:
            out.start_strong()
        self._to_string(out)
        if self.emphasis:
            out.end_emphasis()
        elif self.strong:
            out.end_strong()

class Heading(Container):

    "A heading."

    def __init__(self, nodes, level, start_extra="", start_pad="", end_pad="", end_extra="",
                 identifier=None):
        Container.__init__(self, nodes)
        self.level = level
        self.start_extra = start_extra
        self.start_pad = start_pad
        self.end_pad = end_pad
        self.end_extra = end_extra
        self.identifier = identifier

    def __repr__(self):
        return "Heading(%r, %d, %r, %r, %r, %r, %r)" % (
            self.nodes, self.level, self.start_extra, self.start_pad,
            self.end_pad, self.end_extra, self.identifier)

    def prettyprint(self, indent=""):
        l = ["%sHeading: level=%d start_extra=%r start_pad=%r end_pad=%r"
             " end_extra=%r identifier=%r" % (
             indent, self.level, self.start_extra, self.start_pad, self.end_pad,
             self.end_extra, self.identifier)]
        return self._prettyprint(l, indent)

    def to_string(self, out):
        out.start_heading(self.level, self.start_extra, self.start_pad, self.identifier)
        self._to_string(out)
        out.end_heading(self.level, self.end_pad, self.end_extra)

class LinkLabel(Container):

    "A link or transclusion label."

    def __repr__(self):
        return "LinkLabel(%r)" % self.nodes

    def prettyprint(self, indent=""):
        l = ["%sLinkLabel" % indent]
        return self._prettyprint(l, indent)

    def to_string(self, out):
        out.link_label(self.nodes)

class LinkParameter(Container):

    "A link or transclusion parameter."

    def __repr__(self):
        return "LinkParameter(%r)" % self.nodes

    def prettyprint(self, indent=""):
        l = ["%sLinkParameter" % indent]
        return self._prettyprint(l, indent)

    def to_string(self, out):
        s = self.text_content()
        t = s.split("=", 1)
        out.link_parameter(t)

class List(Container):

    "A list."

    def __init__(self, nodes):
        Container.__init__(self, nodes)
        self.init()

    def init(self):
        self.first = first = self.nodes and self.nodes[0] or None
        self.indent = first and first.indent
        self.marker = first and first.marker
        self.num = first and first.num

    def __repr__(self):
        return "List(%r)" % self.nodes

    def prettyprint(self, indent=""):
        if not self.first:
            self.init()
        l = ["%sList: indent=%r marker=%r num=%r" % (indent, self.indent, self.marker, self.num)]
        return self._prettyprint(l, indent)

    def to_string(self, out):
        if not self.first:
            self.init()
        out.start_list(self.indent, self.marker, self.num)
        self._to_string(out)
        out.end_list(self.indent, self.marker, self.num)

class ListItem(Container):

    "A list item."

    def __init__(self, nodes, indent, marker, space, num):
        Container.__init__(self, nodes)
        self.indent = indent
        self.marker = marker
        self.space = space
        self.num = num

        # Forbid blocks within list items for simpler structure.

        self.allow_blocks = False

    def __repr__(self):
        return "ListItem(%r, %r, %r, %r, %r)" % (self.nodes, self.indent, self.marker, self.space, self.num)

    def prettyprint(self, indent=""):
        l = ["%sListItem: indent=%d marker=%r space=%r num=%r" % (indent, self.indent, self.marker, self.space, self.num)]
        return self._prettyprint(l, indent)

    def to_string(self, out):
        out.start_listitem(self.indent, self.marker, self.space, self.num)
        self._to_string(out)
        out.end_listitem(self.indent, self.marker, self.space, self.num)

class TableAttrs(Container):

    "A collection of table attributes."

    def __repr__(self):
        return "TableAttrs(%r)" % self.nodes

    def prettyprint(self, indent=""):
        l = ["%sTableAttrs:" % indent]
        return self._prettyprint(l, indent)

    def to_string(self, out):
        out.start_table_attrs()
        out.table_attrs(self.nodes)
        out.end_table_attrs()

class Table(Container):

    "A table."

    def __repr__(self):
        return "Table(%r)" % self.nodes

    def prettyprint(self, indent=""):
        l = ["%sTable:" % indent]
        return self._prettyprint(l, indent)

    def to_string(self, out):
        out.start_table()
        self._to_string(out)
        out.end_table()

class TableCell(Container):

    "A table cell."

    def __init__(self, nodes, attrs=None):
        Container.__init__(self, nodes)
        self.attrs = attrs

    def __repr__(self):
        return "TableCell(%r, %r)" % (self.nodes, self.attrs)

    def prettyprint(self, indent=""):
        l = ["%sTableCell:" % indent]
        return self._prettyprint(l, indent)

    def to_string(self, out):
        out.start_table_cell(self.attrs)
        self._to_string(out)
        out.end_table_cell()

class TableRow(Container):

    "A table row."

    def __init__(self, nodes, trailing=""):
        Container.__init__(self, nodes)
        self.trailing = trailing

    def __repr__(self):
        return "TableRow(%r, %r)" % (self.nodes, self.trailing)

    def prettyprint(self, indent=""):
        l = ["%sTableRow: trailing=%r" % (indent, self.trailing)]
        return self._prettyprint(l, indent)

    def to_string(self, out):
        out.start_table_row()
        self._to_string(out)
        out.end_table_row(self.trailing)



# Inline nodes with children.

class Inline(Container):

    "Generic inline formatting."

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.nodes)

    def prettyprint(self, indent=""):
        l = ["%s%s" % (indent, self.__class__.__name__)]
        return self._prettyprint(l, indent)

class Larger(Inline):

    "Larger text."

    def to_string(self, out):
        out.start_larger()
        self._to_string(out)
        out.end_larger()

class Link(Container):

    "Link details."

    def __init__(self, nodes, target):
        Container.__init__(self, nodes)
        self.target = target

    def __repr__(self):
        return "Link(%r, %r)" % (self.nodes, self.target)

    def prettyprint(self, indent=""):
        l = ["%sLink: target=%r" % (indent, self.target)]
        return self._prettyprint(l, indent)

    def to_string(self, out):
        out.link(self.target, self.nodes)

class Macro(Container):

    "Macro details."

    def __init__(self, name, args, parent, region, nodes=None, inline=True):
        Container.__init__(self, nodes or [])
        self.name = name
        self.args = args
        self.parent = parent
        self.region = region
        self.inline = inline

    def __repr__(self):
        return "Macro(%r, %r, %r, %r, %r, %r)" % (self.name, self.args,
                                                  self.parent, self.region,
                                                  self.nodes, self.inline)

    def prettyprint(self, indent=""):
        l = ["%sMacro: name=%r args=%r" % (indent, self.name, self.args)]
        return self._prettyprint(l, indent)

    def to_string(self, out):
        out.start_macro(self.name, self.args, self.nodes, self.inline)
        if self.nodes:
            self._to_string(out)
        out.end_macro(self.inline)

class Monospace(Inline):

    "Monospaced text."

    def to_string(self, out):
        out.start_monospace()
        self._to_string(out)
        out.end_monospace()

class Smaller(Inline):

    "Smaller text."

    def to_string(self, out):
        out.start_smaller()
        self._to_string(out)
        out.end_smaller()

class Strikethrough(Inline):

    "Crossed-out text."

    def to_string(self, out):
        out.start_strikethrough()
        self._to_string(out)
        out.end_strikethrough()

class Subscript(Inline):

    "Subscripted text."

    def to_string(self, out):
        out.start_subscript()
        self._to_string(out)
        out.end_subscript()

class Superscript(Inline):

    "Superscripted text."

    def to_string(self, out):
        out.start_superscript()
        self._to_string(out)
        out.end_superscript()

class Transclusion(Container):

    "Transclusion details."

    def __init__(self, nodes, target):
        Container.__init__(self, nodes)
        self.target = target

    def __repr__(self):
        return "Transclusion(%r, %r)" % (self.nodes, self.target)

    def prettyprint(self, indent=""):
        l = ["%sTransclusion: target=%r" % (indent, self.target)]
        return self._prettyprint(l, indent)

    def to_string(self, out):
        out.transclusion(self.target, self.nodes)

class Underline(Inline):

    "Underlined text."

    def to_string(self, out):
        out.start_underline()
        self._to_string(out)
        out.end_underline()



# Nodes without children.

class Node:

    "A document node without children."

    def empty(self):
        return False

class Anchor(Node):

    "Anchor details."

    def __init__(self, target):
        self.target = target

    def __repr__(self):
        return "Anchor(%r)" % self.target

    def prettyprint(self, indent=""):
        return "%sAnchor: target=%r" % (indent, self.target)

    def to_string(self, out):
        out.anchor(self.target)

class Break(Node):

    "A paragraph break."

    def __repr__(self):
        return "Break()"

    def prettyprint(self, indent=""):
        return "%sBreak" % indent

    def to_string(self, out):
        out.break_()

class Comment(Node):

    "A comment."

    def __init__(self, comment, extra):
        self.comment = comment
        self.extra = extra

    def __repr__(self):
        return "Comment(%r, %r)" % (self.comment, self.extra)

    def prettyprint(self, indent=""):
        return "%sComment: comment=%r extra=%r" % (indent, self.comment, self.extra)

    def to_string(self, out):
        out.comment(self.comment, self.extra)

class Directive(Node):

    "A processing directive."

    def __init__(self, directive, extra):
        self.directive = directive
        self.extra = extra

    def __repr__(self):
        return "Directive(%r, %r)" % (self.directive, self.extra)

    def prettyprint(self, indent=""):
        return "%sDirective: directive=%r extra=%r" % (indent, self.directive, self.extra)

    def to_string(self, out):
        out.directive(self.directive, self.extra)

class LineBreak(Node):

    "A line break within a block."

    def __repr__(self):
        return "LineBreak()"

    def prettyprint(self, indent=""):
        return "%sLineBreak" % indent

    def to_string(self, out):
        out.linebreak()

class Rule(Node):

    "A horizontal rule."

    def __init__(self, height):
        self.height = height

    def __repr__(self):
        return "Rule(%d)" % self.height

    def prettyprint(self, indent=""):
        return "%sRule: height=%d" % (indent, self.height)

    def to_string(self, out):
        out.rule(self.height)

class TableAttr(Node):

    "A table attribute."

    def __init__(self, name, value=None, concise=False, quote=None):
        self.name = name
        self.value = value
        self.concise = concise
        self.quote = quote

    def __repr__(self):
        return "TableAttr(%r, %r, %r, %r)" % (self.name, self.value, self.concise, self.quote)

    def prettyprint(self, indent=""):
        return "%sTableAttr: name=%r value=%r concise=%r quote=%r" % (indent, self.name, self.value, self.concise, self.quote)

    def to_string(self, out):
        out.table_attr(self.name, self.value, self.concise, self.quote)

class Text(Node):

    "A text node."

    def __init__(self, s):
        self.s = s

    def empty(self):
        return not self.s

    def multiline(self):
        return "\n" in self.s

    def merge(self, text):
        self.s += text.s

    def __repr__(self):
        return "Text(%r)" % self.s

    def prettyprint(self, indent=""):
        return "%sText: %r" % (indent, self.s)

    def to_string(self, out):
        out.text(self.s)

class Verbatim(Node):

    "Verbatim text."

    def __init__(self, text):
        self.text = text

    def __repr__(self):
        return "Verbatim(%r)" % self.text

    def prettyprint(self, indent=""):
        return "%sVerbatim: text=%r" % (indent, self.text)

    def to_string(self, out):
        out.verbatim(self.text)

# vim: tabstop=4 expandtab shiftwidth=4
