#!/usr/bin/env python

"""
Directory context functionality.

Copyright (C) 2018, 2019 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from os import makedirs, rename, walk
from os.path import abspath, commonprefix, exists, isdir, isfile, join
import fnmatch

# Get the directory with trailing path separator when assessing path prefixes
# in order to prevent sibling directory confusion.

def inside(filename, dirname):

    "Return whether 'filename' is inside 'dirname'."

    dirprefix = join(dirname, "")
    return filename == dirname or commonprefix((filename, dirprefix)) == dirprefix

def within(filename, dirname):

    "Return the part of 'filename' found within 'dirname'."

    dirname = join(dirname, "")
    prefix = commonprefix((filename, dirname))

    if prefix == dirname:
        return filename[len(prefix):]
    else:
        return None



class Directory:

    "A directory abstraction."

    def __init__(self, filename):

        "Initialise the abstraction with the given 'filename'."

        self.filename = abspath(filename)

    def get_filename(self, filename):

        """
        Return the full path of a file with the given 'filename' found within
        the directory. The full path is an absolute path.
        """

        # Get the absolute path for the combination of directory and filename.

        pathname = abspath(join(self.filename, filename))

        if inside(pathname, self.filename):
            return pathname
        else:
            raise ValueError, filename

    # File operations acting on relative filenames.

    def _apply(self, fn, filename):

        "Apply 'fn' to the relative 'filename'."

        return fn(self.get_filename(filename))

    def ensure(self, filename=None):

        """
        Ensure that this directory, or a directory 'filename' within it, exists.
        """

        pathname = filename and self.get_filename(filename) or self.filename

        if not exists(pathname):
            makedirs(pathname)

    def exists(self, filename):

        "Return whether the relative 'filename' exists within the directory."

        return self._apply(exists, filename)

    def isdir(self, filename):

        "Return whether the relative 'filename' is a directory."

        return self._apply(isdir, filename)

    def isfile(self, filename):

        "Return whether the relative 'filename' is a file."

        return self._apply(isfile, filename)

    def makedirs(self, filename):

        """
        Ensure that a directory having the given 'filename' exists by creating
        it and any directories needed for it to be created. This filename is
        relative to the directory.
        """

        pathname = self.get_filename(filename)

        if not exists(pathname):
            makedirs(pathname)

    def rename(self, old, new):

        """
        Rename the file with the 'old' relative filename to the 'new' relative
        filename.
        """

        rename(self.get_filename(old), self.get_filename(new))

    def select_files(self, pattern, recursive=False):

        """
        Return a list of filenames found within the directory matching
        'pattern'. These filenames are relative to the directory. If 'recursive'
        is specified and is a true value, subdirectories are also searched.
        """

        selected = []

        # Obtain pathnames, directory names and filenames within the directory.

        for dirpath, dirnames, filenames in walk(self.filename):
            if not recursive and dirpath != self.filename:
                continue

            for filename in filenames:

                # Qualify filenames with the directory path.

                pathname = join(dirpath, filename)

                # Obtain the local filename within the directory.

                local_filename = self.within(pathname)

                # Match filenames supporting the pattern.

                if local_filename and fnmatch.fnmatch(local_filename, pattern):
                    selected.append(local_filename)

        return selected

    # File operations involving complete filenames.

    def within(self, filename):

        """
        Return the given complete 'filename' translated to be relative to this
        directory, or return None if the filename describes a location outside
        the directory.
        """

        return within(filename, self.filename)

# vim: tabstop=4 expandtab shiftwidth=4
